/*
 * Copyright (c) 2010-2024. Institut Pasteur.
 *
 * This file is part of Icy.
 * Icy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Icy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Icy. If not, see <https://www.gnu.org/licenses/>.
 */

package plugins.fab.waveletspotdetectorblock;

import org.bioimageanalysis.icy.extension.plugin.abstract_.Plugin;
import org.bioimageanalysis.icy.extension.plugin.annotation_.IcyPluginIcon;
import org.bioimageanalysis.icy.extension.plugin.annotation_.IcyPluginName;
import org.bioimageanalysis.icy.model.roi.ROI;
import org.bioimageanalysis.icy.model.roi.ROI2D;
import org.bioimageanalysis.icy.model.sequence.Sequence;
import org.bioimageanalysis.icy.model.sequence.SequenceUtil;
import org.bioimageanalysis.icy.system.logging.IcyLogger;
import org.bioimageanalysis.icy.system.thread.ThreadUtil;
import org.jetbrains.annotations.NotNull;
import plugins.adufour.blocks.lang.Block;
import plugins.adufour.blocks.util.VarList;
import plugins.adufour.vars.lang.*;
import plugins.adufour.vars.util.VarException;
import plugins.fab.spotDetector.DetectionSpot;
import plugins.fab.spotDetector.detector.DetectionToROI;
import plugins.fab.spotDetector.detector.UDWTScale;
import plugins.fab.spotDetector.detector.UDWTWaveletCore;

import java.util.ArrayList;

@IcyPluginName("Wavelet Spot Detector")
@IcyPluginIcon(path = "/spot-detector-block.png")
public class WaveletSpotDetectorBlock extends Plugin implements Block {
    enum detectionMethodEnum {
        BRIGHT_SPOT, DARK_SPOT
    }

    VarSequence inputSequenceVar = new VarSequence("Input Sequence", null);
    VarEnum<detectionMethodEnum> detectionMethod = new VarEnum<>("Spot to detect", detectionMethodEnum.BRIGHT_SPOT);
    VarBoolean computeWATConsideringROIBoolean = new VarBoolean("Compute WAT considering ROI", false);
    VarSequence inputSequenceROIVar = new VarSequence("ROIs from sequence for WAT", null);
    VarBoolean force2DWaveletVar = new VarBoolean("Force 2D Wavelets", false);
    VarROIArray inputROIForMask = new VarROIArray("ROIs for detection mask", null);

    VarBoolean[] useScaleBoolean = {
            new VarBoolean("Scale 1", false),
            new VarBoolean("Scale 2", true),
            new VarBoolean("Scale 3", false),
            new VarBoolean("Scale 4", false),
            new VarBoolean("Scale 5", false)
    };

    VarInteger[] scaleParameter = {
            new VarInteger("Th. scale 1:", 100),
            new VarInteger("Th. scale 2:", 100),
            new VarInteger("Th. scale 3:", 100),
            new VarInteger("Th. scale 4:", 100),
            new VarInteger("Th. scale 5:", 100)
    };

    //VarSequence outSequence = new VarSequence("output Sequence", null);
    //VarSequence denoisedSequence = new VarSequence("denoised Sequence", null);
    VarSequence binarySpotSequence = new VarSequence("binary Spot Sequence", null);
    VarROIArray detectionAsROIOut = new VarROIArray("detection as ROIs", null);

    @Override
    public void run() {
        // test if input sequence has only 1 channel.

        Sequence inputSequence = inputSequenceVar.getValue();

        if (inputSequence == null) {
            throw new VarException(inputSequenceVar, "The input sequence must be set");
        }

        if (inputSequence.getSizeC() != 1) {
            throw new VarException(inputSequenceVar, "The input sequence must contain only 1 channel. Please use 'Extract Channel' prior to the WaveletBlock");
        }

        // inputChannelVar

        // init detection.

        UDWTWaveletCore waveletCore = new UDWTWaveletCore();

        boolean detectNegative;

        detectNegative = detectionMethod.getValue() != detectionMethodEnum.BRIGHT_SPOT;

        boolean computeWATwithROI = computeWATConsideringROIBoolean.getValue();

        // copy ROI from input sequence to the computation sequence
        Sequence inputComputationSequence;

        try {
            inputComputationSequence = SequenceUtil.getCopy(inputSequence);
        }
        catch (InterruptedException ex) {
            // interrupt process
            return;
        }

        if (inputSequenceROIVar.getValue() != null) {
            for (ROI2D roi : inputSequenceROIVar.getValue().getROI2Ds()) {
                ROI2D roiCopy = (ROI2D) roi.getCopy();

                if (roiCopy != null) {
                    inputComputationSequence.addROI(roiCopy);
                }
            }
        }

        // Icy.addSequence( inputComputationSequence );

        // create UDWTScale Parameter list.

        final ArrayList<UDWTScale> UDWTScaleArrayList = new ArrayList<>();

        // System.out.println("Block debug.");

        ThreadUtil.invokeNow(() -> {
            for (int i = 0; i < 5; i++) {
                // System.out.println( "Scale enabled #" +i +" : " + useScaleBoolean[i].getValue() );
                UDWTScale scale = new UDWTScale(i + 1, useScaleBoolean[i].getValue(), scaleParameter[i].getValue());
                UDWTScaleArrayList.add(scale);
            }
        });

        // Perform detection.

        try {
            ArrayList<DetectionSpot> detectionResult = waveletCore.computeDetection(
                    true,
                    UDWTScaleArrayList,
                    inputComputationSequence,
                    detectNegative,
                    computeWATwithROI,
                    force2DWaveletVar.getValue()
            );

            // System.out.println( "ROI Mask : " + inputROIForMask.getValue() );

            // Filter spot with ROIs.
            if (inputROIForMask.getValue().length != 0) {
                ArrayList<DetectionSpot> detectionResultCopy = new ArrayList<>(detectionResult);

                detectionList:
                for (DetectionSpot spot : detectionResultCopy) {
                    // Point2D spotCoordinates = new Point2D.Double( spot.getMassCenter().x , spot.getMassCenter().y );
                    // spot.getMassCenter()

                    ROI detectionAsROI = null;

                    try {
                        ArrayList<DetectionSpot> detectionList = new ArrayList<>();
                        detectionList.add(spot);
                        detectionAsROI = DetectionToROI.convertDetectionToROI(detectionList).getFirst();
                    }
                    catch (NullPointerException e) {
                        IcyLogger.error(this.getClass(), "Error in spot to ROI conversion.");
                    }

                    if (detectionAsROI == null) {
                        IcyLogger.error(this.getClass(), "Can't convert spot to ROI. spot=" + spot);
                        continue;
                    }

                    // System.out.println( "spot as ROI to test: " + detectionAsROI.getBounds5D() );

                    for (ROI roi : inputROIForMask.getValue()) {
                        // System.out.println( "roi contenant: " + roi.getBounds5D() );

                        if (roi.contains(detectionAsROI)) {
                            // System.out.println("contains");
                            continue detectionList;
                        }
                    }
                    // System.out.println("don't contains");

                    detectionResult.remove(spot);

                }
            }

            // set outputs.

            // binary output: propagate ROIs.
            {
                Sequence binarySequence = waveletCore.getBinarySequence();

                if (inputSequenceROIVar.getValue() != null) {
                    for (ROI2D roi : inputSequenceROIVar.getValue().getROI2Ds()) {
                        // System.out.println("roi:"+roi);
                        String name = roi.getName();

                        ROI2D roiCopy = (ROI2D) roi.getCopy();
                        roi.setName(name); // because of the bug when coppying roi.
                        roiCopy.setName(name); // else the name has " copy" in the end of it.

                        binarySequence.addROI(roiCopy);
                    }
                }

                binarySpotSequence.setValue(binarySequence);
            }

            // output detection as ROI
            {
                ArrayList<ROI> roiList = DetectionToROI.convertDetectionToROI(detectionResult);
                ROI[] roi = new ROI[roiList.size()];
                int i = 0;
                for (ROI roiCandidate : roiList) {
                    roi[i] = roiCandidate;
                    i++;
                }
                detectionAsROIOut.setValue(roi);
            }
        }
        catch (InterruptedException e) {
            // protocols process interrupted
            Thread.currentThread().interrupt();
        }
    }

    @Override
    public void declareInput(@NotNull VarList inputMap) {
        inputMap.add("input-sequence", inputSequenceVar);
        inputMap.add("detection-method", detectionMethod);
        inputMap.add("input-sequence-rois", inputSequenceROIVar);
        inputMap.add("compute-wat-considering-roi", computeWATConsideringROIBoolean);
        inputMap.add("input-roi-for-mask", inputROIForMask);

        for (int i = 0; i < 5; i++) {
            inputMap.add("use-scale-" + i, useScaleBoolean[i]);
        }
        for (int i = 0; i < 5; i++) {
            inputMap.add("scale-value-" + i, scaleParameter[i]);
        }
        inputMap.add("force-2d-wavelet", force2DWaveletVar);
    }

    @Override
    public void declareOutput(@NotNull VarList outputMap) {
        // outputMap.add( outSequence );
        // outputMap.add( denoisedSequence );
        outputMap.add("Binary Spot Sequence", binarySpotSequence);
        outputMap.add("Detection as ROI", detectionAsROIOut);

    }

}
